﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;

namespace RecrownedAthenaeum.Camera
{
    /// <summary>
    /// A virtual 2D camera.
    /// </summary>
    public class Camera2D
    {
        /// <summary>
        /// Current zoom level.
        /// </summary>
        public float Zoom;

        /// <summary>
        /// Current position in the world.
        /// </summary>
        public Vector2 Position;

        /// <summary>
        /// The matrix representing the zoom and position.
        /// </summary>
        public Matrix Matrix { get; private set; }

        private GraphicsDevice graphicsDevice;

        /// <summary>
        /// Constructs 2D camera.
        /// </summary>
        /// <param name="graphicsDevice">The graphics device to use.</param>
        public Camera2D(GraphicsDevice graphicsDevice)
        {
            this.graphicsDevice = graphicsDevice;
            Zoom = 1f;
            Position.X = this.graphicsDevice.Viewport.Width * 0.5f;
            Position.Y = this.graphicsDevice.Viewport.Height * 0.5f;
        }

        /// <summary>
        /// Applies any changes made to the properties of this camera and updates to new dimensions of viewport specified by the graphics device.
        /// </summary>
        public void Update()
        {
            Rectangle bounds = graphicsDevice.Viewport.Bounds;
            Matrix = 
                Matrix.CreateTranslation(new Vector3(-Position.X, -Position.Y, 0)) * 
                Matrix.CreateScale(Zoom) * 
                Matrix.CreateTranslation(new Vector3(bounds.Width * 0.5f, bounds.Height * 0.5f, 0f));
        }

        /// <summary>
        /// Lerps to the given position.
        /// </summary>
        /// <param name="alpha">The multiplier for difference in distance.</param>
        /// <param name="targetPosition">The target position to lerp to.</param>
        /// <param name="delta">Time between this frame and the previous frame.</param>
        public void LinearInterpolationToPosition(float alpha, Vector2 targetPosition, float delta)
        {
            if (alpha <= 0 && alpha > 1f) throw new ArgumentException("Alpha can't be greater than 1f, less than or equal to 0.");

            Vector2 distance = targetPosition - Position;
            distance *= (float)(1.0f - Math.Pow(1 - alpha, delta / 0.02f));

            Position += distance;
        }
    }
}
